/*
 * llx - parallel command invoker
 *
 * Copyright (C) 2011-2012 Ali Gholami Rudi
 *
 * This program is released under GNU GPL version 2.
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>

#define MAXARGS		(1 << 10)

static void forkall(char *s)
{
	char *argv[MAXARGS];
	char *e = strchr(s, '\n');
	int argc = 0;
	if (!fork()) {
		if (strchr(s, '|') || strchr(s, '"') || strchr(s, '`')) {
			argv[argc++] = "sh";
			argv[argc++] = "-c";
			argv[argc++] = s;
			*e = '\0';
		} else {
			while (s < e) {
				char *a = s;
				while (!isspace(*s))
					s++;
				*s++ = '\0';
				argv[argc++] = a;
				while (s < e && isspace(*s))
					s++;
			}
		}
		execvp(argv[0], argv);
		fprintf(stderr, "executing <%s> failed\n", argv[0]);
		exit(1);
	}
}

static char cmds[1 << 18];

int xread(int fd, char *buf, int len)
{
	int nr = 0;
	while (nr < len) {
		int ret = read(fd, buf + nr, len - nr);
		if (ret <= 0)
			break;
		nr += ret;
	}
	return nr;
}

int main(int argc, char *argv[])
{
	int nr;
	char *s, *e, *r;
	int n = 0;
	int maxn = 1 << 10;
	nr = xread(0, cmds, sizeof(cmds));
	s = cmds;
	e = cmds + nr;
	if (argc > 1 && isdigit(argv[1][0]))
		maxn = atoi(argv[1]);
	while (s < e) {
		r = memchr(s, '\n', e - s);
		if (!r)
			break;
		forkall(s);
		s = r + 1;
		n++;
		while (n >= maxn) {
			waitpid(-1, NULL, 0);
			n--;
		}
	}
	while (n--)
		waitpid(-1, NULL, 0);
	return 0;
}
